#cython: language_level=3
import datetime
import decimal
import os
import uuid
import cerializer.constants

from cpython.int cimport PyInt_AS_LONG
from cpython.tuple cimport PyTuple_GET_ITEM
from libc.time cimport mktime, tm


if cerializer.constants.QUANTLANE:
	import qutils.time.nanotime



# Mostly taken from fastavro version 0.22.6
# Original file: https://github.com/fastavro/fastavro/blob/master/fastavro/_logical_writers.pyx


'''
This module deals with preparation of logical types.
It is only used from within Cerializer.
'''



ctypedef long long long64

cdef is_windows = os.name == 'nt'
cdef has_timestamp_fn = hasattr(datetime.datetime, 'timestamp')

cdef long64 MCS_PER_SECOND = 1000000
cdef long64 MCS_PER_MINUTE = 60000000
cdef long64 MCS_PER_HOUR = 3600000000

cdef long64 MLS_PER_SECOND = 1000
cdef long64 MLS_PER_MINUTE = 60000
cdef long64 MLS_PER_HOUR = 3600000


epoch = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
epoch_naive = datetime.datetime(1970, 1, 1)
DAYS_SHIFT = datetime.date(1970, 1, 1).toordinal()



cdef inline mk_bits(bits):
	return bytes([bits & 0xff])


cdef inline prepare_timestamp_millis(object data):
	cdef object tt
	cdef tm time_tuple
	if isinstance(data, datetime.datetime):
		if not has_timestamp_fn:
			if data.tzinfo is not None:
				return <long64>(<double>(
					<object>(data - epoch).total_seconds()) * MLS_PER_SECOND
				)
			tt = data.timetuple()
			time_tuple.tm_sec = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 5)))
			time_tuple.tm_min = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 4)))
			time_tuple.tm_hour = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 3)))
			time_tuple.tm_mday = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 2)))
			time_tuple.tm_mon = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 1))) - 1
			time_tuple.tm_year = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 0))) - 1900
			time_tuple.tm_isdst = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 8)))
			return mktime(& time_tuple) * MLS_PER_SECOND + <long64>(
				int(data.microsecond) / 1000)
		else:
			# On Windows, timestamps before the epoch will raise an error.
			# See https://bugs.python.org/issue36439
			if is_windows:
				if data.tzinfo is not None:
					return <long64>(<double>(
						<object>(data - epoch).total_seconds()) * MLS_PER_SECOND
					)
				else:
					return <long64>(<double>(
						<object>(data - epoch_naive).total_seconds()) * MLS_PER_SECOND
					)
			else:
				x =  <long64>(<double>(data.timestamp()) * MLS_PER_SECOND)
				return x
	else:
		return data


cdef inline prepare_timestamp_micros(object data):
	cdef object tt
	cdef tm time_tuple
	if isinstance(data, datetime.datetime):
		if not has_timestamp_fn:
			if data.tzinfo is not None:
				return <long64>(<double>(
					<object>(data - epoch).total_seconds()) * MCS_PER_SECOND
				)
			tt = data.timetuple()
			time_tuple.tm_sec = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 5)))
			time_tuple.tm_min = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 4)))
			time_tuple.tm_hour = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 3)))
			time_tuple.tm_mday = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 2)))
			time_tuple.tm_mon = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 1))) - 1
			time_tuple.tm_year = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 0))) - 1900
			time_tuple.tm_isdst = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 8)))

			return mktime(& time_tuple) * MCS_PER_SECOND + \
				<long64>(data.microsecond)
		else:
			# On Windows, timestamps before the epoch will raise an error.
			# See https://bugs.python.org/issue36439
			if is_windows:
				if data.tzinfo is not None:
					return <long64>(<double>(
						<object>(data - epoch).total_seconds()) * MCS_PER_SECOND
					)
				else:
					return <long64>(<double>(
						<object>(data - epoch_naive).total_seconds()) * MCS_PER_SECOND
					)
			else:
				return <long64>(<double>(data.timestamp()) * MCS_PER_SECOND)
	else:
		return data


cdef inline prepare_date(object data):
	if isinstance(data, datetime.date):
		return data.toordinal() - DAYS_SHIFT
	elif isinstance(data, str):
		return datetime.datetime.strptime(data, "%Y-%m-%d").toordinal() - DAYS_SHIFT
	else:
		return data


cdef inline prepare_bytes_decimal(object data, schema):
	"""Convert decimal.Decimal to bytes"""
	if not isinstance(data, decimal.Decimal):
		return data
	scale = schema.get('scale', 0)

	sign, digits, exp = data.as_tuple()

	delta = exp + scale

	if delta < 0:
		raise ValueError(
			'Scale provided in schema does not match the decimal')

	unscaled_datum = 0
	for digit in digits:
		unscaled_datum = (unscaled_datum * 10) + digit

	unscaled_datum = 10 ** delta * unscaled_datum

	bytes_req = (unscaled_datum.bit_length() + 8) // 8

	if sign:
		unscaled_datum = -unscaled_datum

	return unscaled_datum.to_bytes(bytes_req, byteorder='big', signed=True)



cdef inline prepare_fixed_decimal(object data, schema):
	cdef bytearray tmp
	if not isinstance(data, decimal.Decimal):
		return data
	scale = schema.get('scale', 0)
	size = schema['size']

	# based on https://github.com/apache/avro/pull/82/

	sign, digits, exp = data.as_tuple()

	if -exp > scale:
		raise ValueError(
			'Scale provided in schema does not match the decimal')
	delta = exp + scale
	if delta > 0:
		digits = digits + (0,) * delta

	unscaled_datum = 0
	for digit in digits:
		unscaled_datum = (unscaled_datum * 10) + digit

	bits_req = unscaled_datum.bit_length() + 1

	size_in_bits = size * 8
	offset_bits = size_in_bits - bits_req

	mask = 2 ** size_in_bits - 1
	bit = 1
	for i in range(bits_req):
		mask ^= bit
		bit <<= 1

	if bits_req < 8:
		bytes_req = 1
	else:
		bytes_req = bits_req // 8
		if bits_req % 8 != 0:
			bytes_req += 1

	tmp = bytearray()

	if sign:
		unscaled_datum = (1 << bits_req) - unscaled_datum
		unscaled_datum = mask | unscaled_datum
		for index in range(size - 1, -1, -1):
			bits_to_write = unscaled_datum >> (8 * index)
			tmp += mk_bits(bits_to_write & 0xff)
	else:
		for i in range(offset_bits // 8):
			tmp += mk_bits(0)
		for index in range(bytes_req - 1, -1, -1):
			bits_to_write = unscaled_datum >> (8 * index)
			tmp += mk_bits(bits_to_write & 0xff)

	return tmp


cdef inline prepare_uuid(object data):
	if isinstance(data, uuid.UUID):
		return str(data)
	else:
		return data



cdef inline prepare_time_millis(object data):
	if isinstance(data, datetime.time):
		return int(
			data.hour * MLS_PER_HOUR + data.minute * MLS_PER_MINUTE
			+ data.second * MLS_PER_SECOND + int(data.microsecond / 1000))
	else:
		return data



cdef inline prepare_time_micros(object data):
	if isinstance(data, datetime.time):
		return int(data.hour * MCS_PER_HOUR + data.minute * MCS_PER_MINUTE
					+ data.second * MCS_PER_SECOND + data.microsecond)
	else:
		return data



cdef prepare_nano_time(data):
	return data.nanoseconds



cdef prepare_nano_time_delta(data):
	return data.nanoseconds



cdef inline read_timestamp_millis(data):
	return parse_timestamp(data, float(1000))

cdef inline read_timestamp_micros(data):
	return parse_timestamp(data, float(1000000))

cdef inline read_date(data):
	return datetime.date.fromordinal(data + datetime.date(1970, 1, 1).toordinal())

cdef inline read_uuid(data):
	return uuid.UUID(data)

cdef inline read_decimal(data, schema):
	scale = schema.get('scale', 0)
	precision = schema['precision']
	unscaled_datum = int.from_bytes(data, byteorder = 'big', signed = True)
	decimal_context = decimal.Context()
	decimal_context.prec = precision
	return decimal_context.create_decimal(unscaled_datum). \
		scaleb(-scale, decimal_context)

cdef inline read_time_millis(data):
	h = int(data / 60 * 60 * 1000)
	m = int(data / 60 * 1000) % 60
	s = int(data / 1000) % 60
	mls = int(data % 1000) * 1000
	return datetime.time(h, m, s, mls)

cdef inline read_time_micros(data):
	h = int(data / (60 * 60 * 1000000))
	m = int(data / (60 * 1000000)) % 60
	s = int(data / 1000000) % 60
	mcs = data % 1000000
	return datetime.time(h, m, s, mcs)

cdef inline parse_timestamp(data, resolution):
	return datetime.datetime(1970, 1, 1, tzinfo = datetime.timezone.utc) + datetime.timedelta(seconds = data / resolution)

cdef read_nano_time(data):
	return qutils.time.nanotime.NanoTime(data)

cdef read_nano_time_delta(data):
	return qutils.time.NanoTimeDelta(data)



# pylint: disable=protected-access
import itertools
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import cerializer.schema_parser


def correct_type(type_: Union[Dict[str, Any], str, List[Any]]) -> Optional[str]:
	'''
	Corrects the nuances between Avro type definitions and actual python type names.
	'''
	if type_ == 'string':
		return 'str'
	if type_ == 'boolean':
		return 'bool'
	if type_ == 'long':
		return 'int'
	if type_ == 'double':
		return 'float'
	if isinstance(type_, str) and type_ in ('int', 'null', 'float', 'bytes'):
		return type_
	return None


def get_logical_type_constraint(schema: Dict[str, Any], location: str) -> str:
	logical_type = schema['logicalType'].replace('-', '_')
	data_type = schema['type']
	if logical_type == 'decimal':
		params = {'scale': schema.get('scale', 0), 'size': schema.get('size', 0)}
		return f'type(prepare.prepare_{data_type}_{logical_type}({location}, {params})) is {data_type}'
	return f'type(prepare.prepare_{logical_type}({location})) is {data_type}'


def name_generator(prefix: str) -> Iterator[str]:
	yield from (f'{prefix}_{i}' for i in itertools.count())


def parse_schema(schema: Union[Dict[str, Any], list, str]) -> Any:
	'''
	Wrapper for loading schemata from yaml
	'''
	while True:
		try:
			parsed = cerializer.schema_parser.parse_schema(schema)
			return parsed
		except cerializer.schema_parser.UnknownType as e:
			# we ignore missing schema errors since we are going to fill them in later
			cerializer.schema_parser.SCHEMA_DEFS[e.name] = {}


def get_subschemata(schemata: List[Tuple[str, Any]]) -> Dict[str, Union[str, List, Dict[str, Any]]]:
	schema_database: Dict[str, Union[str, List, Dict[str, Any]]] = {}
	for schema_identifier, schema in schemata:
		parsed_schema = parse_schema(schema)
		if '.' in schema_identifier:
			schema_database[schema_identifier] = cerializer.schema_parser.parse_schema(parsed_schema)
		scan_schema_for_subschemata(parsed_schema, schema_database)
	return schema_database


def scan_schema_for_subschemata(schema: Any, schema_database: Dict[str, Any]) -> None:
	if type(schema) is dict:
		name = schema.get('name')
		if name and '.' in name:
			schema_database[name] = schema
		for _, subschema in schema.items():
			scan_schema_for_subschemata(subschema, schema_database)
	if type(schema) in (list, dict):
		for subschema in schema:
			scan_schema_for_subschemata(subschema, schema_database)


def get_type_name(type_: Union[str, Dict[str, Any]]) -> Optional[str]:
	return type_ if isinstance(type_, str) else type_.get('name')


def default_if_necessary(location: str, default: Any) -> str:
	if default is None:
		return ''
	default = 'None' if default == 'null' else default
	if '[' not in location:
		constraint = f'if {location} is None:'
	else:
		r = location.rfind(']')
		l = location.rfind('[')
		constraint = f'if {location[:l]}.get({location[l+1:r]}) is None:'
	return f'{constraint}\n' f'    {location} = {default}'


def get_schema_identifier(namespace: str, schema_name: str) -> str:
	return f'{namespace}.{schema_name}'



import io
from typing import Any, Dict, List, Union

import cerializer.compiler
import cerializer.schema_handler
import cerializer.utils


class Cerializer:
	'''
	The main entry point of Cerializer project.
	A Cerializer instance renders and then compiles code for given schema.
	It then provides two methods - serialize and deserialize.
	These methods are then called by the user to de/serialize data.
	'''

	def __init__(
		self,
		cerializer_schemata: cerializer.schema_handler.CerializerSchemata,
		namespace: str,
		schema_name: str,
	) -> None:
		'''
		One can initialize Cerializer with a list of schema_identifier, schema or with a list of schema roots.
		For schema roots usage checkout README.
		'''
		self.schema_identifier = cerializer.utils.get_schema_identifier(namespace, schema_name)
		self.code_generator = cerializer.schema_handler.CodeGenerator(cerializer_schemata, self.schema_identifier,)
		self.cerializer_schemata = cerializer_schemata
		compiled_code = self._get_compiled_code(
			self.cerializer_schemata.load_schema(self.schema_identifier, self.schema_identifier)
		)
		self._serialization_function = compiled_code['serialize']
		self._deserialization_function = compiled_code['deserialize']

	def deserialize(self, data: bytes) -> Any:
		data_io = io.BytesIO(data)
		return self._deserialization_function(data_io)

	def deserialize_from(self, buffer: io.BytesIO) -> Any:
		return self._deserialization_function(buffer)

	def serialize(self, data: Any) -> bytes:  # the result is stored in the output variable
		output = io.BytesIO()
		self._serialization_function(data, output)
		return output.getvalue()

	def serialize_into(self, buffer: io.BytesIO, data: Any) -> None:
		self._serialization_function(data, buffer)

	def _get_compiled_code(self, schema: Union[str, List[Any], Dict[str, Any]]) -> Any:
		code = self.code_generator.render_code_with_wraparounds(schema)
		return cerializer.compiler.compile_code(code)



# pylint: disable=protected-access, deprecated-method, no-value-for-parameter
import distutils.core
import hashlib
import importlib.machinery
import os.path
import sys
from typing import Any, List

import Cython.Build.Dependencies
import Cython.Build.Inline
import Cython.Compiler.Main
import Cython.Utils


def compile_code(code: str) -> Any:
	return _cython_inline(code)


def _load_dynamic(name: str, module_path: str) -> Any:
	# mypy does not understand ExtensionFileLoader init
	return importlib.machinery.ExtensionFileLoader(name, module_path).load_module()  # type: ignore


def _cython_inline(
	complete_code: str,
	lib_dir: str = os.path.join(Cython.Utils.get_cython_cache_dir(), 'inline'),
) -> Any:
	key = complete_code, sys.version_info, sys.executable, 3, Cython.__version__
	module_name = '_cython_inline_' + hashlib.md5(str(key).encode('utf-8')).hexdigest()

	build_extension = Cython.Build.Inline._get_build_extension()
	module_path = os.path.join(lib_dir, module_name + build_extension.get_ext_filename(''))

	code = []
	for line in complete_code.split('\n'):
		code.append('    ' + line)

	if not os.path.exists(lib_dir):
		os.makedirs(lib_dir)
	cflags: List[str] = []
	c_include_dirs: List[str] = []
	pyx_file = os.path.join(lib_dir, module_name + '.pyx')
	fh = open(pyx_file, 'w')
	try:
		fh.write(complete_code)
	finally:
		fh.close()
	extension = distutils.core.Extension(
		name = module_name,
		sources = [pyx_file],
		include_dirs = c_include_dirs,
		extra_compile_args = cflags,
	)
	if build_extension is None:
		build_extension = Cython.Build.Inline._get_build_extension()
	build_extension.extensions = Cython.Build.Dependencies.cythonize(
		[extension],
		include_path = ['.'],
		quiet = True,
	)
	build_extension.build_temp = os.path.dirname(pyx_file)
	build_extension.build_lib = lib_dir
	build_extension.run()

	module = _load_dynamic(module_name, module_path)

	return module.__invoke()



import os
import timeit

import yaml

import cerializer.cerializer_handler
import cerializer.schema_handler
import cerializer.tests.test_cerializer


def schemata() -> cerializer.schema_handler.CerializerSchemata:
	schemata = []
	for schema_identifier, schema_root in cerializer.tests.test_cerializer.iterate_over_schemata():
		# mypy things yaml has no attribute unsafe_load, which is not true
		schema_tuple = (
			schema_identifier,
			yaml.unsafe_load(open(os.path.join(schema_root, 'schema.yaml'))),  # type: ignore
		)
		schemata.append(schema_tuple)
	return cerializer.schema_handler.CerializerSchemata(schemata)


def benchmark(number: int) -> None:
	results = []
	total_cerializer = 0.0
	total_fastavro = 0.0
	for schema_identifier, path in cerializer.tests.test_cerializer.iterate_over_schemata():
		setup = f'''
import cerializer.tests.test_cerializer as t
import cerializer.tests.benchmark as b
import yaml
import cerializer.cerializer_handler
import fastavro
import os
import io


schema_identifier = '{schema_identifier}'
CERIALIZER_SCHEMATA = b.schemata()
cerializer_codec = cerializer.cerializer_handler.Cerializer(
	cerializer_schemata = CERIALIZER_SCHEMATA,
	namespace = schema_identifier.split('.')[0],
	schema_name = schema_identifier.split('.')[1]
)
data = list(yaml.unsafe_load_all(open(os.path.join('{path}', 'example.yaml'))))[0]  # type: ignore
SCHEMA_FAVRO = fastavro.parse_schema(
	yaml.load(open(os.path.join('{path}', 'schema.yaml')), Loader = yaml.Loader)
)
		'''

		stmt_cerializer = '''
output_cerializer = cerializer_codec.serialize(data)
deserialized_data = cerializer_codec.deserialize(output_cerializer)
		'''

		stmt_fastavro = '''
output_fastavro = io.BytesIO()
fastavro.schemaless_writer(output_fastavro, SCHEMA_FAVRO, data)
x = output_fastavro.getvalue()
output_fastavro.seek(0)
res = fastavro.schemaless_reader(output_fastavro, SCHEMA_FAVRO)
		'''

		result_cerializer = timeit.timeit(stmt = stmt_cerializer, setup = setup, number = number)
		result_fastavro = timeit.timeit(stmt = stmt_fastavro, setup = setup, number = number)
		total_cerializer += result_cerializer
		total_fastavro += result_fastavro
		results.append(
			f'{schema_identifier.ljust(36, " ")}   {(result_fastavro/result_cerializer):.4f} times faster,   {result_fastavro:.4f}s : {result_cerializer:.4f}s'
		)
	for r in results:
		print(r)  # dumb_style_checker:disable = print-statement
	print(  # dumb_style_checker:disable = print-statement
		f'AVERAGE: {(total_fastavro/total_cerializer):.4f} times faster'
	)



# pylint: disable=protected-access
import copy
import os
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import jinja2

import cerializer.utils
from cerializer import constants


class CerializerSchemata:
	'''
	Storage class for schemata.
	'''

	def __init__(self, schemata: List[Tuple[str, Any]]):
		self._schema_database: Dict[str, Union[str, List[Any], Dict[str, Any]]] = cerializer.utils.get_subschemata(
			schemata
		)
		self._cycle_starting_nodes: Set[str] = set()
		self._init_cycles()

	def __contains__(self, item: str) -> bool:
		return item in self._schema_database

	def load_schema(
		self,
		schema_identifier: str,
		context_schema_identifier: Optional[str] = None,
	) -> Union[str, List, Dict[str, Any]]:
		# we first check whether the schema we are looking for is not defined in the same big schema
		# this would mean the schema is redefined and that the local version has to be used
		if context_schema_identifier:
			context_schema = self._schema_database[context_schema_identifier]
			# mypy needs this type annotation
			local_schema_database: Dict[str, Union[str, List[Any], Dict[str, Any]]] = cerializer.utils.get_subschemata(
				[(context_schema_identifier, context_schema)]
			)
			if schema_identifier in local_schema_database:
				return local_schema_database[schema_identifier]
		if schema_identifier in self._schema_database:
			return self._schema_database[schema_identifier]
		else:
			raise RuntimeError(f'Schema with identifier = {schema_identifier} not found in schema database.')

	def is_cycle_starting(self, schema_identifier: str) -> bool:
		return schema_identifier in self._cycle_starting_nodes

	def _init_cycles(self) -> None:
		for _, schema in self._schema_database.items():
			visited: Set[str] = set()
			self._cycle_detection(schema, visited, self._cycle_starting_nodes)

	def _cycle_detection(
		self,
		parsed_schema: Union[str, List, Dict[str, Any]],
		visited: Set[str],
		cycle_starting_nodes: Set[str],
	) -> None:
		'''
		Detects cycles in schemata.
		This can happen when for example a schema is defined using itself eg. a tree schema.
		This method add all cycle starting nodes in all schemata_database to cycle_starting_nodes set.
		'''
		if isinstance(parsed_schema, str) and parsed_schema in visited:
			cycle_starting_nodes.add(parsed_schema)
		elif isinstance(parsed_schema, dict):
			name = parsed_schema.get('name')
			type_ = parsed_schema['type']
			if type(type_) is str and type_ in visited:
				cycle_starting_nodes.add(type_)
			elif name:
				visited.add(name)
				new_visited = copy.deepcopy(visited)
				if 'fields' in parsed_schema:
					for field in parsed_schema['fields']:
						self._cycle_detection(field, new_visited, cycle_starting_nodes)
				if type(type_) is dict:
					self._cycle_detection(type_, new_visited, cycle_starting_nodes)
				if type(type_) is list:
					for element in type_:
						self._cycle_detection(element, new_visited, cycle_starting_nodes)
				elif type(type_) is str and type_ in self:
					self._cycle_detection(self.load_schema(type_), new_visited, cycle_starting_nodes)


class CodeGenerator:
	'''
	Driver class for code generation.
	'''

	def __init__(self, schemata: CerializerSchemata, schema_identifier: str) -> None:
		self.context_schema = schema_identifier
		self.buffer_name = 'buffer'
		self.cdefs: List[str] = []
		self.schemata = schemata
		self.dict_name_generator = cerializer.utils.name_generator('d_dict')
		self.val_name_generator = cerializer.utils.name_generator('val')
		self.key_name_generator = cerializer.utils.name_generator('key')
		self.type_name_generator = cerializer.utils.name_generator('type')
		self.int_name_generator = cerializer.utils.name_generator('i')
		self.jinja_env = jinja2.Environment(
			loader = jinja2.FileSystemLoader(
				searchpath = os.path.join(constants.PROJECT_ROOT, 'cerializer', 'templates')
			),
		)
		self.jinja_env.globals['env'] = self.jinja_env
		# this is a bool flag for turning on the DictWrapper feature
		self.jinja_env.globals['quantlane'] = constants.QUANTLANE
		self.necessary_defs: Set[str] = set()
		self.handled_cycles: Set[str] = set()

	def prepare(
		self,
		mode: constants.SerializationMode,
		logical_type: str,
		data_type: str,
		location: str,
		schema: Dict[str, Any],
	) -> str:
		'''
		Returns a de/serialization function string for logical types that need to be prepared first.
		Will raise RuntimeError
		'''
		logical_type = logical_type.replace('-', '_')

		if 'nano_time' in logical_type and not constants.QUANTLANE:
			raise RuntimeError('Trying to serialize Nano Time logical type without the flag quantlane being True. ')
		params = {
			'scale': schema.get('scale', 0),
			'size': schema.get('size', 0),
			'precision': schema.get('precision'),
		}

		read_params = f'fo, {params}' if data_type == 'fixed' else 'fo'
		if mode is constants.SerializationMode.MODE_SERIALIZE:
			prepare_params = f'{location}, {params}' if logical_type == 'decimal' else location
			prepare_type = f'{data_type}_{logical_type}' if logical_type == 'decimal' else logical_type
			return f'write.write_{data_type}({self.buffer_name}, prepare.prepare_{prepare_type}({prepare_params}))'
		else:
			prepare_params = (
				f'read.read_{data_type}({read_params}), {params}'
				if logical_type == 'decimal'
				else f'read.read_{data_type}({read_params})'
			)
			return f'{location} = prepare.read_{logical_type}({prepare_params})'

	def get_serialization_function(self, type_: str, location: str) -> str:
		'''
		Returns the corresponding serialization function call string.
		'''
		if type_ == 'null':
			return f'write.write_null({self.buffer_name})'
		if type_ in self.schemata and type_ not in constants.BASIC_TYPES:
			schema = self.load_with_context(type_)
			old_context = self.context_schema
			self.context_schema = type_
			code = self.generate_serialization_code(schema, location)
			self.context_schema = old_context
			return code
		return f'write.write_{type_}({self.buffer_name}, {location})'

	def get_deserialization_function(
		self,
		type_: str,
		location: str,
		schema: Optional[Dict[str, Any]] = None,
	) -> str:
		'''
		Returns the corresponding deserialization function call string.
		'''
		if type_ == 'null':
			return f'{location} = None'
		if type_ in self.schemata and type_ not in constants.BASIC_TYPES:
			loaded_schema = self.load_with_context(type_)
			old_context = self.context_schema
			self.context_schema = type_
			code = self.generate_deserialization_code(loaded_schema, location)
			self.context_schema = old_context
			return code
		read_params = f'fo, {schema}' if type_ == 'fixed' else 'fo'
		return f'{location} = read.read_{type_}({read_params})'

	def get_union_index_function(self, index: int) -> str:
		'''
		Returns a function call string for union index.
		'''
		return f'write.write_long({self.buffer_name}, {index})'

	def get_array_serialization(self, schema: Dict[str, Any], location: str) -> str:
		'''
		Return array serialization string.
		'''
		item_name = next(self.val_name_generator)
		item_serialization_code = self.generate_serialization_code(schema['items'], item_name)
		template = self.jinja_env.get_template('array_serialization.jinja2')
		return template.render(
			location = location,
			buffer_name = self.buffer_name,
			item_serialization_code = item_serialization_code,
			item_name = item_name,
		)

	def get_array_deserialization(self, schema: Dict[str, Any], location: str) -> str:
		'''
		Return array deserialization string.
		'''
		index_name = next(self.int_name_generator)
		block_count_name = next(self.int_name_generator)
		potential_item_name = next(self.val_name_generator)
		self.add_cdef('long long', index_name)
		self.add_cdef('long long', block_count_name)
		template = self.jinja_env.get_template('array_deserialization.jinja2')
		return template.render(
			location = location,
			buffer_name = self.buffer_name,
			items = schema['items'],
			index_name = index_name,
			block_count_name = block_count_name,
			potential_item_name = potential_item_name,
		)

	def get_enum_serialization(self, schema: Dict[str, Any], location: str) -> str:
		'''
		Return enum serialization string.
		'''
		symbols = schema['symbols']
		return f'write.write_int({self.buffer_name}, {symbols}.index({location}))'

	def get_enum_deserialization(self, schema: Dict[str, Any], location: str) -> str:
		'''
		Return enum deserialization string.
		'''
		symbols = schema['symbols']
		return f'{location} = {symbols}[read.read_int(fo)]'

	def get_union_serialization(
		self,
		schema: Union[List, Dict[str, Any]],
		location: str,
		is_from_array: bool = False,
	) -> str:
		'''
		Return union serialization string.
		'''
		if is_from_array:
			# this is in case a union is not specified as a standalone type but is declared in array items
			type_ = list(schema)  # since this union schema came from an array, it has to be in list form
			name = None
			new_location = location
		elif isinstance(schema, dict):
			name = schema['name']
			type_ = list(schema['type'])  # schema['type'] has to be list since its a union schema
			new_location = f"{location}['{name}']"

		else:
			raise NotImplementedError(f'Cant handle schema = {schema}')

		if len([item for item in type_ if (isinstance(item, dict) and item.get('type') == 'array')]) > 1:
			# this is documented in task CL-132
			raise NotImplementedError(
				'One of your schemas contains a union of more than one array types. This is not yet implemented.'
			)

		possible_types_and_code = []
		# we need to ensure that null is checked first
		if 'null' in type_:
			possible_types_and_code.append(
				(
					'null',
					self.get_union_index_function(type_.index('null')),
					self.get_serialization_function('null', new_location),
				)
			)
		for possible_type in type_:
			if possible_type == 'null':
				continue
			possible_types_and_code.append(
				(
					possible_type,
					self.get_union_index_function(type_.index(possible_type)),
					self.generate_serialization_code(possible_type, new_location),
				)
			)
		type_name = next(self.type_name_generator)
		self.add_cdef('str', type_name)
		data_name = next(self.val_name_generator)
		template = self.jinja_env.get_template('union_serialization.jinja2')
		if is_from_array:
			return template.render(
				types = possible_types_and_code,
				location = location,
				name = name,
				type_name = type_name,
				data_name = data_name,
				value = location,
			)
		return template.render(
			types = possible_types_and_code,
			location = location,
			name = name,
			type_name = type_name,
			data_name = data_name,
		)

	def get_union_deserialization(
		self,
		schema: Union[List, Dict[str, Any]],
		location: str,
		is_from_array: bool = False,
	) -> str:
		'''
		Return union serialization string.
		'''
		index_name = next(self.int_name_generator)
		self.add_cdef('long', index_name)
		if is_from_array:
			types = schema
			new_location = location
		elif isinstance(schema, dict):
			name = schema['name']
			types = schema['type']
			new_location = f"{location}['{name}']"
		else:
			raise NotImplementedError(f'Cant handle schema = {schema}')
		template = self.jinja_env.get_template('union_deserialization.jinja2')
		return template.render(index_name = index_name, types = types, location = new_location,)

	def get_map_serialization(self, schema: Dict[str, Any], location: str) -> str:
		'''
		Return map serialization string.
		'''
		dict_name = next(self.dict_name_generator)
		self.add_cdef('dict', dict_name)
		template = self.jinja_env.get_template('map_serialization.jinja2')
		values = schema['values']
		key_name = next(self.key_name_generator)
		val_name = next(self.val_name_generator)
		self.add_cdef('str', key_name)
		return template.render(
			location = location,
			buffer_name = self.buffer_name,
			values = values,
			key_name = key_name,
			val_name = val_name,
		)

	def get_map_deserialization(self, schema: Dict[str, Any], location: str) -> str:
		'''
		Return map deserialization string.
		'''
		key_name = next(self.key_name_generator)
		self.add_cdef('unicode', key_name)
		block_count_name = next(self.int_name_generator)
		self.add_cdef('long', block_count_name)
		index_name = next(self.int_name_generator)
		self.add_cdef('long', index_name)

		template = self.jinja_env.get_template('map_deserialization.jinja2')
		values = schema['values']
		return template.render(
			location = location,
			values = values,
			key_name = key_name,
			block_count_name = block_count_name,
			index_name = index_name,
		)

	def render_code_with_wraparounds(self, schema: Union[str, List[Any], Dict[str, Any]]) -> str:
		code = self.render_code(schema = schema)
		meta_template = self.jinja_env.get_template('meta_template.jinja2')
		rendered_code = meta_template.render(code = code,)
		return rendered_code

	def render_code(self, schema: Union[str, List, Dict[str, Any]]) -> str:
		'''
		Renders code for a given schema into a .pyx file.
		'''
		self.jinja_env.globals['correct_type'] = cerializer.utils.correct_type
		self.jinja_env.globals['correct_constraint'] = self.correct_constraint
		self.jinja_env.globals['generate_serialization_code'] = self.generate_serialization_code
		self.jinja_env.globals['generate_deserialization_code'] = self.generate_deserialization_code
		self.jinja_env.globals['get_type_name'] = cerializer.utils.get_type_name
		schema = cerializer.utils.parse_schema(schema)
		location = 'data'
		serialization_code = self.generate_serialization_code(schema = schema, location = location)
		serialization_code = '\n'.join(self.cdefs) + '\n' + serialization_code
		self.cdefs = []
		deserialization_code = self.generate_deserialization_code(schema = schema, location = location)
		deserialization_code = '\n'.join(self.cdefs) + '\n' + deserialization_code

		template = self.jinja_env.get_template('template.jinja2')
		rendered_body = template.render(
			location = location,
			buffer_name = self.buffer_name,
			serialization_code = serialization_code,
			deserialization_code = deserialization_code,
			necessary_defs = '\n\n\n\n'.join([i for i in self.necessary_defs if i != '']),
		)
		self.cdefs = []
		self.necessary_defs = set()
		return rendered_body

	def generate_serialization_code(self, schema: Union[str, List, Dict[str, Any]], location: str) -> str:
		'''
		Driver function to handle code generation for a schema.
		'''
		if isinstance(schema, str):
			if self.schemata.is_cycle_starting(schema) and schema not in constants.BASIC_TYPES:
				return self.handle_cycle(constants.SerializationMode.MODE_SERIALIZE, schema, location)
			return self.get_serialization_function(schema, location)
		if isinstance(schema, list):
			return self.get_union_serialization(schema, location, is_from_array = True)
		type_ = schema['type']
		if 'logicalType' in schema:
			prepared = self.prepare(
				constants.SerializationMode.MODE_SERIALIZE,
				schema['logicalType'],
				type_,
				location,
				schema,
			)
			return prepared
		elif type_ == constants.RECORD:
			return '\n'.join((self.generate_serialization_code(field, location)) for field in schema['fields'])
		elif type_ == constants.ARRAY:
			return self.get_array_serialization(schema, location)
		elif type_ == constants.ENUM:
			return self.get_enum_serialization(schema, location)
		elif type_ == constants.MAP:
			return self.get_map_serialization(schema, location)
		elif type_ == constants.FIXED:
			return self.get_serialization_function(type_, location)
		elif type(type_) is dict:
			name = schema['name']
			new_location = f"{location}['{name}']"
			default_if_necessary = cerializer.utils.default_if_necessary(new_location, schema.get('default'))
			default_if_necessary = (default_if_necessary + '\n') if default_if_necessary else ''
			return str(default_if_necessary + self.generate_serialization_code(type_, new_location))
		elif type(type_) is list:
			return self.get_union_serialization(schema, location)
		elif type(type_) is str and type_ in constants.BASIC_TYPES:
			name = schema.get('name')
			if name:
				location = f"{location}['{name}']"
			default_if_necessary = cerializer.utils.default_if_necessary(location, schema.get('default'))
			default_if_necessary = (default_if_necessary + '\n') if default_if_necessary else ''
			return str(default_if_necessary + self.get_serialization_function(type_, location))
		elif type(type_) is str and type_ in self.schemata:
			loaded_schema = self.load_with_context(type_)
			old_context = self.context_schema
			self.context_schema = type_
			if self.schemata.is_cycle_starting(type_):
				return self.handle_cycle(constants.SerializationMode.MODE_SERIALIZE, type_, location)
			name = schema['name']
			new_location = f"{location}['{name}']"
			code = self.generate_serialization_code(loaded_schema, new_location)
			self.context_schema = old_context
			return code
		raise NotImplementedError(f'Cant handle schema = {schema}')

	def load_with_context(self, schema_identifier: str) -> Union[str, List, Dict[str, Any]]:
		return self.schemata.load_schema(schema_identifier, self.context_schema)

	def generate_deserialization_code(self, schema: Union[Dict[str, Any], list, str], location: str) -> str:
		'''
		Driver function to handle code generation for a schema.
		'''
		if isinstance(schema, str):
			if self.schemata.is_cycle_starting(schema) and schema not in constants.BASIC_TYPES:
				return self.handle_cycle(constants.SerializationMode.MODE_DESERIALIZE, schema, location)
			return self.get_deserialization_function(schema, location)
		if isinstance(schema, list):
			return self.get_union_deserialization(schema, location, is_from_array = True)
		if isinstance(schema, dict):
			type_ = schema['type']
			if 'logicalType' in schema:
				prepared = self.prepare(
					constants.SerializationMode.MODE_DESERIALIZE,
					schema['logicalType'],
					type_,
					location,
					schema,
				)
				return prepared
			elif type_ == constants.RECORD:
				field_deserialization = '\n'.join(
					(self.generate_deserialization_code(field, location))
					for field in schema['fields']
				)
				return location + ' = {}\n' + field_deserialization
			elif type_ == constants.ARRAY:
				return self.get_array_deserialization(schema, location)
			elif type_ == constants.ENUM:
				return self.get_enum_deserialization(schema, location)
			elif type_ == constants.MAP:
				return self.get_map_deserialization(schema, location)
			elif type_ == constants.FIXED:
				return self.get_deserialization_function(type_, location, schema = schema)
			elif type(type_) is dict:
				name = schema['name']
				new_location = f"{location}['{name}']"
				return self.generate_deserialization_code(type_, new_location)
			elif type(type_) is list:
				return self.get_union_deserialization(schema, location)
			elif type(type_) is str and type_ in constants.BASIC_TYPES:
				name = schema.get('name')
				if name:
					location = f"{location}['{name}']"
				return self.get_deserialization_function(type_, location, schema = schema)
			elif type(type_) is str and type_ in self.schemata:
				loaded_schema = self.load_with_context(type_)
				old_context = self.context_schema
				self.context_schema = type_
				if self.schemata.is_cycle_starting(type_):
					return self.handle_cycle(constants.SerializationMode.MODE_DESERIALIZE, type_, location)
				name = schema['name']
				new_location = f"{location}['{name}']"
				code = self.generate_deserialization_code(loaded_schema, new_location)
				self.context_schema = old_context
				return code
		raise NotImplementedError(f'Cant handle schema = {schema}')

	def handle_cycle(self, mode: constants.SerializationMode, schema: str, location: str) -> str:
		normalised_type = schema.replace(':', '_').replace('.', '_')
		serialization_function = (
			f'{constants.SerializationMode.MODE_SERIALIZE.value}_{normalised_type}(data, output)'
		)
		deserialization_function = f'{constants.SerializationMode.MODE_DESERIALIZE.value}_{normalised_type}(fo)'
		if schema not in self.handled_cycles:
			self.handled_cycles.add(schema)
			code = self.render_code(self.load_with_context(schema))
			self.necessary_defs.add(
				code.replace(
					f'cpdef {constants.SerializationMode.MODE_SERIALIZE.value}(data, output)',
					f'def {serialization_function}',
				)
				.replace(
					f'def {constants.SerializationMode.MODE_SERIALIZE.value}(data, output)',
					f'def {serialization_function}',
				)
				.replace(
					f'cpdef {constants.SerializationMode.MODE_DESERIALIZE.value}(fo)',
					f'def {deserialization_function}',
				)
				.replace(
					f'def {constants.SerializationMode.MODE_DESERIALIZE.value}(fo)',
					f'def {deserialization_function}',
				)
			)
		serialization_function_call = serialization_function.replace('(data,', f'({location},')
		if mode is constants.SerializationMode.MODE_SERIALIZE:
			return f'output.write(buffer)\nbuffer = bytearray()\n{serialization_function_call}'
		else:
			return f'{location} = {deserialization_function}'

	def add_cdef(self, type_: str, name: str) -> None:
		cdef = f'cdef {type_} {name}'
		self.cdefs.append(cdef)

	def correct_constraint(
		self,
		type_: Union[Dict[str, Any], str, List],
		location: str,
		key: str,
		first: bool,
		value: Optional[str] = None,
	) -> str:
		# value is filled when we are passing in a name of a local variable rather then a dict and a string
		if value:
			full_location = value
		else:
			full_location = f'{location}["{key}"]'
		correct_type_ = cerializer.utils.correct_type(type_)
		constraint = None

		if correct_type_:
			if correct_type_ == 'null':
				if value:
					constraint = f'{full_location} is None'
				else:
					constraint = f'"{key}" not in {location} or {location}["{key}"] is None'
			else:
				constraint = f'type({full_location}) is {cerializer.utils.correct_type(type_)}'

		elif isinstance(type_, dict) and type_.get('type') == 'fixed':
			constraint = f'type({full_location}) is bytes'

		elif isinstance(type_, dict) and type_.get('type') == 'array':
			constraint = f'type({full_location}) is list'

		elif isinstance(type_, dict) and type_.get('type') == 'map':
			constraint = f'type({full_location}) is dict'

		elif isinstance(type_, dict) and type_.get('type') == 'enum':
			constraint = f'type({full_location}) is str and {full_location} in {type_["symbols"]}'

		elif isinstance(type_, dict) and type_.get('logicalType') is not None:
			constraint = cerializer.utils.get_logical_type_constraint(type_, full_location)

		elif isinstance(type_, str) and type_ in self.schemata:
			return self.correct_constraint(self.load_with_context(type_), location, key, first, value)

		elif isinstance(type_, dict) and type_['type'] == 'record':
			# TODO adjust for different dict types
			constraint = f'type({full_location}) is dict'

		if constraint:
			return f'{"if" if first else "elif"} {constraint}:'
		raise RuntimeError(f'invalid constraint for type == {type_}')



# pylint: disable=protected-access
import io
import logging
import os
from typing import Iterator, Tuple

import fastavro
import pytest
import yaml

import cerializer.cerializer_handler
import cerializer.schema_handler
import cerializer.schema_parser
import cerializer.utils



SCHEMA_ROOTS = [os.path.join(os.path.dirname(__file__), 'schemata')]


def iterate_over_schemata() -> Iterator[Tuple[str, str]]:
	for schema_root in SCHEMA_ROOTS:
		for schema_identifier in os.listdir(schema_root):
			if schema_identifier.startswith('.'):
				# in case of folders automatically added by macOS (.DS_Store)
				continue
			yield schema_identifier, os.path.join(schema_root, schema_identifier)


@pytest.fixture(scope = 'module')
def schemata() -> cerializer.schema_handler.CerializerSchemata:
	schemata = []
	for schema_identifier, schema_root in iterate_over_schemata():
		# mypy things yaml has no attribute unsafe_load, which is not true
		schema_tuple = schema_identifier, yaml.unsafe_load( # type: ignore
			open(os.path.join(schema_root, 'schema.yaml'))
		)
		schemata.append(schema_tuple)
	return cerializer.schema_handler.CerializerSchemata(schemata)


@pytest.mark.parametrize(
	'schema_identifier, schema_root',
	iterate_over_schemata(),
)
@pytest.mark.parametrize('use_serialize_into', (
	True,
	False,
))
def test_fastavro_compatibility_serialize(
	schema_root: str,
	schema_identifier: str,
	schemata: cerializer.schema_handler.CerializerSchemata,
	use_serialize_into: bool,
) -> None:
	# patch for not working avro codec
	namespace = schema_identifier.split('.')[0]
	schema_name = schema_identifier.split('.')[1]
	cerializer_codec = cerializer.cerializer_handler.Cerializer(
		cerializer_schemata = schemata,
		namespace = namespace,
		schema_name = schema_name,
	)
	try:
		# mypy things yaml has no attribute unsafe_load_all, which is not true
		data_all = yaml.unsafe_load_all( # type: ignore
			open(os.path.join(schema_root, 'example.yaml')))
		SCHEMA_FAVRO = fastavro.parse_schema(
			yaml.load(
				open(os.path.join(schema_root, 'schema.yaml')), Loader = yaml.Loader
			)
		)
		for data in data_all:
			output_fastavro = io.BytesIO()
			fastavro.schemaless_writer(output_fastavro, SCHEMA_FAVRO, data)
			if use_serialize_into:
				buffer = io.BytesIO()
				cerializer_codec.serialize_into(buffer, data)
				output_cerializer = buffer.getvalue()
			else:
				output_cerializer = cerializer_codec.serialize(data)
			assert output_cerializer == output_fastavro.getvalue()
	except FileNotFoundError:
		logging.warning('Missing schema or Example file for schema == %s', schema_name)
		assert False


@pytest.mark.parametrize(
	'schema_identifier, schema_root',
	iterate_over_schemata(),
)
@pytest.mark.parametrize('use_deserialize_from', (
	True,
	False,
))
def test_fastavro_compatibility_deserialize(
	schema_root: str,
	schema_identifier: str,
	schemata: cerializer.schema_handler.CerializerSchemata,
	use_deserialize_from: bool,
) -> None:
	# patch for not working avro codec
	namespace = schema_identifier.split('.')[0]
	schema_name = schema_identifier.split('.')[1]
	cerializer_codec = cerializer.cerializer_handler.Cerializer(
		cerializer_schemata = schemata,
		namespace = namespace,
		schema_name = schema_name,
	)
	try:
		# mypy things yaml has no attribute unsafe_load_all, which is not true
		data_all = yaml.unsafe_load_all( # type: ignore
			open(os.path.join(schema_root, 'example.yaml'))
		)
		SCHEMA_FAVRO = yaml.load(
			open(os.path.join(schema_root, 'schema.yaml')), Loader = yaml.Loader
		)
		for data in data_all:
			output_fastavro = io.BytesIO()
			fastavro.schemaless_writer(output_fastavro, SCHEMA_FAVRO, data)
			output_fastavro.seek(0)
			if use_deserialize_from:
				deserialized = cerializer_codec.deserialize_from(output_fastavro)
			else:
				deserialized = cerializer_codec.deserialize(output_fastavro.getvalue())

			output_fastavro.seek(0)
			assert deserialized == fastavro.schemaless_reader(output_fastavro, SCHEMA_FAVRO)
	except FileNotFoundError:
		logging.warning(
			'Missing schema or Example file for schema == %s',
			schema_name,
		)
		assert False



#cython: language_level=3
ctypedef int int32
ctypedef unsigned int uint32
ctypedef unsigned long long ulong64
ctypedef long long long64
cdef union double_ulong64:
	double d
	ulong64 n



cpdef inline read_null(fo):
	"""null is written as zero bytes."""
	return None


cpdef inline read_boolean(fo):
	"""A boolean is written as a single byte whose value is either 0 (false) or
	1 (true).
	"""
	cdef unsigned char ch_temp
	cdef bytes bytes_temp = fo.read(1)
	# technically 0x01 == true and 0x00 == false, but many languages will
	# cast anything other than 0 to True and only 0 to False
	ch_temp = bytes_temp[0]
	return ch_temp != 0


cpdef inline long64 read_long(fo) except? -1:
	return read_int(fo)


cpdef inline long64 read_int(fo) except? -1:
	"""int and long values are written using variable-length, zig-zag
	coding."""
	cdef ulong64 b
	cdef ulong64 n
	cdef int32 shift

	cdef bytes c = fo.read(1)

	# We do EOF checking only here, since most reader start here
	if not c:
		raise StopIteration

	b = <unsigned char>(c[0])
	n = b & 0x7F
	shift = 7

	while (b & 0x80) != 0:
		c = fo.read(1)
		b = <unsigned char>(c[0])
		n |= (b & 0x7F) << shift
		shift += 7

	return (n >> 1) ^ -(n & 1)


cdef union float_uint32:
	float f
	uint32 n


cpdef inline read_float(fo):
	"""A float is written as 4 bytes.

	The float is converted into a 32-bit integer using a method equivalent to
	Java's floatToIntBits and then encoded in little-endian format.
	"""
	cdef bytes data
	cdef unsigned char ch_data[4]
	cdef float_uint32 fi
	data = fo.read(4)
	ch_data[:4] = data
	fi.n = (ch_data[0]
			| (ch_data[1] << 8)
			| (ch_data[2] << 16)
			| (ch_data[3] << 24))
	return fi.f


cdef union double_ulong64:
	double d
	ulong64 n


cpdef inline read_double(fo):
	"""A double is written as 8 bytes.

	The double is converted into a 64-bit integer using a method equivalent to
	Java's doubleToLongBits and then encoded in little-endian format.
	"""
	cdef bytes data
	cdef unsigned char ch_data[8]
	cdef double_ulong64 dl
	data = fo.read(8)
	ch_data[:8] = data
	dl.n = (ch_data[0]
			| (<ulong64>(ch_data[1]) << 8)
			| (<ulong64>(ch_data[2]) << 16)
			| (<ulong64>(ch_data[3]) << 24)
			| (<ulong64>(ch_data[4]) << 32)
			| (<ulong64>(ch_data[5]) << 40)
			| (<ulong64>(ch_data[6]) << 48)
			| (<ulong64>(ch_data[7]) << 56))
	return dl.d



cpdef inline read_bytes(fo):
	"""Bytes are encoded as a long followed by that many bytes of data."""
	cdef long64 size = read_int(fo)
	return fo.read(<long>size)


cpdef inline unicode read_string(fo):
	"""A string is encoded as a long followed by that many bytes of UTF-8
	encoded character data.
	"""
	return read_bytes(fo).decode('utf-8')


cpdef inline read_fixed(fo, writer_schema):
	"""Fixed instances are encoded using the number of bytes declared in the
	schema."""
	return fo.read(writer_schema['size'])



# Taken from Fastavro version 0.22.6 due to technical import reasons
# Original file: https://github.com/fastavro/fastavro/blob/master/fastavro/_logical_writers.pyx
from typing import Any, Dict, List, Optional, Set, Tuple, Union


PRIMITIVES = {
	'boolean',
	'bytes',
	'double',
	'float',
	'int',
	'long',
	'null',
	'string',
}

# A mapping of named schemas to their actual schema definition
SCHEMA_DEFS: Dict[str, Union[Dict[str, Any], str, List]] = {}

RESERVED_PROPERTIES = {
	'type',
	'name',
	'namespace',
	'fields',  # Record
	'items',  # Array
	'size',  # Fixed
	'symbols',  # Enum
	'values',  # Map
	'doc',
}

OPTIONAL_FIELD_PROPERTIES = {
	'doc',
	'aliases',
	'default',
}

RESERVED_FIELD_PROPERTIES = {'type', 'name'} | OPTIONAL_FIELD_PROPERTIES


class UnknownType(ValueError):
	def __init__(self, name: str) -> None:
		super(UnknownType, self).__init__(name)
		self.name = name


class SchemaParseException(Exception):
	pass


def parse_schema(
	schema: Union[Dict[str, Any], str, List],
	expand: bool = False,
	_write_hint: bool = True,
	_force: bool = False,
) -> Union[Dict[str, Any], str, List]:
	if _force or expand:
		return _parse_schema(schema, '', expand, _write_hint, set())
	elif isinstance(schema, dict) and '__fastavro_parsed' in schema:
		return schema
	elif isinstance(schema, list):
		# If we are given a list we should make sure that the immediate sub
		# schemas have the hint in them
		return [parse_schema(s, expand, _write_hint, _force) for s in schema]
	else:
		return _parse_schema(schema, '', expand, _write_hint, set())


def _parse_schema(
	schema: Union[Dict[str, Any], str, List],
	namespace: str,
	expand: bool,
	_write_hint: bool,
	named_schemas: Set[str],
) -> Union[Dict[str, Any], str, List]:
	# union schemas
	if isinstance(schema, list):
		return [_parse_schema(s, namespace, expand, False, named_schemas) for s in schema]

	# string schemas; this could be either a named schema or a primitive type
	elif not isinstance(schema, dict):
		if schema in PRIMITIVES:
			return schema

		if '.' not in schema and namespace:
			schema = namespace + '.' + schema

		if schema not in SCHEMA_DEFS:
			raise UnknownType(schema)
		elif expand:
			# If `name` is in the schema, it has been fully resolved and so we
			# can include the full schema. If `name` is not in the schema yet,
			# then we are still recursing that schema and must use the named
			# schema or else we will have infinite recursion when printing the
			# final schema
			if 'name' in SCHEMA_DEFS[schema]:
				return SCHEMA_DEFS[schema]
			else:
				return schema
		else:
			return schema

	else:
		# Remaining valid schemas must be dict types
		schema_type = schema['type']

		parsed_schema = {key: value for key, value in schema.items() if key not in RESERVED_PROPERTIES}
		parsed_schema['type'] = schema_type

		if 'doc' in schema:
			parsed_schema['doc'] = schema['doc']

		# Correctness checks for logical types
		logical_type = parsed_schema.get('logicalType')
		if logical_type == 'decimal':
			scale = parsed_schema.get('scale')
			if scale and not isinstance(scale, int):
				raise SchemaParseException('decimal scale must be a postive integer, ' + 'not {}'.format(scale))
			precision = parsed_schema.get('precision')
			if precision and not isinstance(precision, int):
				raise SchemaParseException('decimal precision must be a postive integer, ' + 'not {}'.format(precision))

		if schema_type == 'array':
			parsed_schema['items'] = _parse_schema(schema['items'], namespace, expand, False, named_schemas,)

		elif schema_type == 'map':
			parsed_schema['values'] = _parse_schema(schema['values'], namespace, expand, False, named_schemas,)

		elif schema_type == 'enum':
			_, fullname = schema_name(schema, namespace)
			if fullname in named_schemas:
				raise SchemaParseException('redefined named type: {}'.format(fullname))
			named_schemas.add(fullname)

			SCHEMA_DEFS[fullname] = parsed_schema

			parsed_schema['name'] = fullname
			parsed_schema['symbols'] = schema['symbols']

		elif schema_type == 'fixed':
			_, fullname = schema_name(schema, namespace)
			if fullname in named_schemas:
				raise SchemaParseException('redefined named type: {}'.format(fullname))
			named_schemas.add(fullname)

			SCHEMA_DEFS[fullname] = parsed_schema

			parsed_schema['name'] = fullname
			parsed_schema['size'] = schema['size']

		elif schema_type == 'record' or schema_type == 'error':
			# records
			namespace, fullname = schema_name(schema, namespace)
			if fullname in named_schemas:
				raise SchemaParseException('redefined named type: {}'.format(fullname))
			named_schemas.add(fullname)

			SCHEMA_DEFS[fullname] = parsed_schema

			fields = []
			for field in schema.get('fields', []):
				fields.append(parse_field(field, namespace, expand, named_schemas))

			parsed_schema['name'] = fullname
			parsed_schema['fields'] = fields

			# Hint that we have parsed the record
			if _write_hint:
				parsed_schema['__fastavro_parsed'] = True

		elif schema_type in PRIMITIVES:
			parsed_schema['type'] = schema_type

		else:
			raise UnknownType(str(schema))

		return parsed_schema


def parse_field(
	field: Dict[str, Any],
	namespace: str,
	expand: bool,
	named_schemas: Set[str],
) -> Dict[str, Any]:
	parsed_field = {key: value for key, value in field.items() if key not in RESERVED_FIELD_PROPERTIES}

	for prop in OPTIONAL_FIELD_PROPERTIES:
		if prop in field:
			parsed_field[prop] = field[prop]

	# Aliases must be a list
	aliases = parsed_field.get('aliases', [])
	if not isinstance(aliases, list):
		msg = 'aliases must be a list, not {}'.format(aliases)
		raise SchemaParseException(msg)

	parsed_field['name'] = field['name']
	parsed_field['type'] = _parse_schema(field['type'], namespace, expand, False, named_schemas)

	return parsed_field


def schema_name(schema: Dict[str, Any], parent_ns: Optional[str]) -> Tuple[str, str]:
	try:
		name = schema['name']
	except KeyError:
		msg = '"name" is a required field missing from ' + 'the schema: {}'.format(schema)
		raise SchemaParseException(msg)

	namespace = schema.get('namespace', parent_ns)
	if not namespace:
		return namespace, name

	return namespace, '{}.{}'.format(namespace, name)



#cython: language_level=3
'''
This module deals with writing basic types. 
Complex types are decomposed first in schema_parser.

This code is heavily inspired by FastAvro https://github.com/fastavro/fastavro
'''


ctypedef int int32
ctypedef unsigned int uint32
ctypedef unsigned long long ulong64
ctypedef long long long64
cdef union double_ulong64:
    double d
    ulong64 n



cdef inline write_null(object fo):
    """null is written as zero bytes"""
    pass


cdef inline write_boolean(bytearray fo, bint datum):
    """A boolean is written as a single byte whose value is either 0 (false) or
    1 (true)."""
    cdef unsigned char ch_temp[1]
    ch_temp[0] = 1 if datum else 0
    fo += ch_temp[:1]


cdef inline write_int(bytearray fo, datum):
    """int and long values are written using variable-length, zig-zag coding.
    """
    cdef ulong64 n
    cdef unsigned char ch_temp[1]
    n = (datum << 1) ^ (datum >> 63)
    while (n & ~0x7F) != 0:
        ch_temp[0] = (n & 0x7f) | 0x80
        fo += ch_temp[:1]
        n >>= 7
    ch_temp[0] = n
    fo += ch_temp[:1]


cdef inline write_long(bytearray fo, datum):
    write_int(fo, datum)


cdef union float_uint32:
    float f
    uint32 n


cdef inline write_float(bytearray fo, float datum):
    """A float is written as 4 bytes.  The float is converted into a 32-bit
    integer using a method equivalent to Java's floatToIntBits and then encoded
    in little-endian format."""
    cdef float_uint32 fi
    cdef unsigned char ch_temp[4]

    fi.f = datum
    ch_temp[0] = fi.n & 0xff
    ch_temp[1] = (fi.n >> 8) & 0xff
    ch_temp[2] = (fi.n >> 16) & 0xff
    ch_temp[3] = (fi.n >> 24) & 0xff

    fo += ch_temp[:4]



cdef inline write_double(bytearray fo, double datum, schema=None):
    """A double is written as 8 bytes.  The double is converted into a 64-bit
    integer using a method equivalent to Java's doubleToLongBits and then
    encoded in little-endian format.  """
    cdef double_ulong64 fi
    cdef unsigned char ch_temp[8]

    fi.d = datum
    ch_temp[0] = fi.n & 0xff
    ch_temp[1] = (fi.n >> 8) & 0xff
    ch_temp[2] = (fi.n >> 16) & 0xff
    ch_temp[3] = (fi.n >> 24) & 0xff
    ch_temp[4] = (fi.n >> 32) & 0xff
    ch_temp[5] = (fi.n >> 40) & 0xff
    ch_temp[6] = (fi.n >> 48) & 0xff
    ch_temp[7] = (fi.n >> 56) & 0xff

    fo += ch_temp[:8]


cdef inline write_bytes(bytearray fo, bytes datum, schema=None):
    """Bytes are encoded as a long followed by that many bytes of data."""
    write_long(fo, len(datum))
    fo += datum


cdef inline write_string(bytearray fo, datum, schema=None):
    """A string is encoded as a long followed by that many bytes of UTF-8
    encoded character data."""
    write_bytes(fo, bytes(datum, 'UTF-8'))



cdef inline write_fixed(bytearray fo, object datum, schema=None):
    """Fixed instances are encoded using the number of bytes declared in the
    schema."""
    fo += datum


cdef inline write_enum(bytearray fo, datum, schema):
    """An enum is encoded by a int, representing the zero-based position of
    the symbol in the schema."""
    index = schema['symbols'].index(datum)
    write_int(fo, index)



import os

from setuptools import Extension, find_packages, setup


COMPILED_MODULES = {
	'prepare.pyx',
}
MODULES_TO_BUILD = []
EXTENSIONS = []

for file_name in COMPILED_MODULES:
	c_file = file_name.replace('.pyx', '.c').replace('.py', '.c')
	if os.path.isfile(c_file):
		EXTENSIONS.append(Extension(os.path.splitext(file_name)[0].replace('/', '.'), sources = [c_file]))
	else:
		MODULES_TO_BUILD.append(file_name)

try:
	from Cython.Build import cythonize

	EXTENSIONS += cythonize(MODULES_TO_BUILD)
except ImportError:
	if len(EXTENSIONS) != len(COMPILED_MODULES):
		raise RuntimeError('Cannot cythonize required modules.')

ROOT = os.path.dirname(__file__)

setup(
	name = 'cerializer',
	author = 'matejmicek.com',
	author_email = 'matej.micek@quantlane.com',
	url = 'https://github.com/matejmicek/Cerializer',
	version = open(os.path.join(ROOT, 'version.txt')).read().strip(),
	packages = find_packages(exclude = ['test*']),
	package_data = {
		'cerializer': [
			'templates/*.jinja2',
			'../write.pxd',
			'../read.pxd',
			'../prepare.pxd',
			'../prepare.pyx',
		],
	},
	install_requires = [
		'Cython>=0.28.4,<1.0.0',
		'PyYAML>=5.3.1,<6.0.0',
		'setuptools>=46.0.0',
		'Jinja2>=2.11.2,<3.0.0',
	],
	extras_require = {
		'quantlane': [
			'ql-qutils>=9.5.2,<10.0.0'
		],
	},
	setup_requires = [
		'Cython>=0.28.4,<1.0.0',
	],
	ext_modules = EXTENSIONS,
)



#cython: language_level=3

cdef mk_bits(bits)

cdef mk_bits(bits)

cdef prepare_timestamp_millis(object data)

cdef prepare_timestamp_micros(object data)

cdef prepare_date(object data)

cdef prepare_bytes_decimal(object data, schema)

cdef prepare_fixed_decimal(object data, schema)

cdef prepare_uuid(object data)

cdef prepare_time_millis(object data)

cdef prepare_time_micros(object data)

cdef prepare_nano_time(data)

cdef prepare_nano_time_delta(data)

cdef read_timestamp_millis(data)

cdef read_timestamp_micros(data)

cdef read_date(data)

cdef read_uuid(data)

cdef read_decimal(data, schema)

cdef read_time_millis(data)

cdef read_time_micros(data)

cdef parse_timestamp(data, resolution)

cdef read_nano_time(data)

cdef read_nano_time_delta(data)



import enum
import importlib.util
import os


BOOLEAN = 'boolean'
INT = 'int'
LONG = 'long'
FLOAT = 'float'
DOUBLE = 'double'
BYTES = 'bytes'
STRING = 'string'

RECORD = 'record'
ENUM = 'enum'
ARRAY = 'array'
MAP = 'map'
FIXED = 'fixed'

BASIC_TYPES = {BOOLEAN, INT, LONG, FLOAT, DOUBLE, BYTES, STRING}

COMPLEX_TYPES = {RECORD, ENUM, ARRAY, MAP, FIXED}

PROJECT_ROOT = os.path.join(os.path.dirname(__file__), '..')


SCHEMA_ROOTS = [os.path.join(PROJECT_ROOT, 'cerializer', 'tests', 'schemata')]


class SerializationMode(enum.Enum):
	MODE_SERIALIZE = 'serialize'
	MODE_DESERIALIZE = 'deserialize'


QUANTLANE = False

package_name = 'qutils'
spec = importlib.util.find_spec(package_name)
if spec:
	# there is no way to know whether the user used the extra 'quantlane'
	# other than searching for the packages
	QUANTLANE = True
